﻿#region License Information
/* HeuristicLab
 * Copyright (C) Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
 #endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Drawing;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HEAL.Attic;

namespace HeuristicLab.DatastreamAnalysis {
  /// <summary>
  /// A table of data values or intervals.
  /// </summary>
  [Item("ExtendedDataTable", "A data table where the points are either double values or ranges")]
  [StorableType("8E716B16-5B9C-441D-97BB-8C362E8D5BC4")]
  public class ExtendedDataTable : NamedItem, IStringConvertibleMatrix, IContent {
    public static new Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Performance; }
    }

    private DataTableVisualProperties visualProperties;
    public DataTableVisualProperties VisualProperties {
      get { return visualProperties; }
      set {
        if (visualProperties != value) {
          if (value == null) throw new ArgumentNullException("VisualProperties");
          if (visualProperties != null) visualProperties.PropertyChanged -= new PropertyChangedEventHandler(VisualProperties_PropertyChanged);
          visualProperties = value;
          visualProperties.PropertyChanged += new PropertyChangedEventHandler(VisualProperties_PropertyChanged);
          OnVisualPropertiesChanged();
        }
      }
    }

    private NamedItemCollection<RangeDataRow> rows;
    public NamedItemCollection<RangeDataRow> Rows {
      get { return rows; }
      private set {
        if (rows != null) throw new InvalidOperationException("Rows already set.");
        rows = value;
        if (rows != null) RegisterRowsEvents();
      }
    }

    #region Persistence Properties
    [Storable(Name = "VisualProperties")]
    private DataTableVisualProperties StorableVisualProperties {
      get { return visualProperties; }
      set {
        visualProperties = value;
        visualProperties.PropertyChanged += new PropertyChangedEventHandler(VisualProperties_PropertyChanged);
      }
    }
    [Storable(Name = "rows")]
    private IEnumerable<RangeDataRow> StorableRows {
      get { return rows; }
      set { rows = new NamedItemCollection<RangeDataRow>(value); }
    }
    #endregion

    #region constructors, cloner,...
    [StorableConstructor]
    protected ExtendedDataTable(StorableConstructorFlag _) : base(_) { }
    protected ExtendedDataTable(ExtendedDataTable original, Cloner cloner) : base(original, cloner) {
      VisualProperties = (DataTableVisualProperties)cloner.Clone(original.visualProperties);
      rows = cloner.Clone(original.rows);
      this.RegisterRowsEvents();
    }
    public ExtendedDataTable() : base() {
      Name = "ExtendedDataTable";
      VisualProperties = new DataTableVisualProperties();
      rows = new NamedItemCollection<RangeDataRow>();
      this.RegisterRowsEvents();
    }
    public ExtendedDataTable(string name) : base(name) {
      VisualProperties = new DataTableVisualProperties();
      rows = new NamedItemCollection<RangeDataRow>();
      this.RegisterRowsEvents();
    }
    public ExtendedDataTable(string name, string description) : base(name, description) {
      VisualProperties = new DataTableVisualProperties();
      rows = new NamedItemCollection<RangeDataRow>();
      this.RegisterRowsEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExtendedDataTable(this, cloner);
    }
    #endregion constructors, cloner,...

    #region events
    public event EventHandler VisualPropertiesChanged;
    protected virtual void OnVisualPropertiesChanged() {
      var handler = VisualPropertiesChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void VisualProperties_PropertyChanged(object sender, PropertyChangedEventArgs e) {
      OnVisualPropertiesChanged();
    }

    private void RegisterRowsEvents() {
      rows.ItemsAdded += new CollectionItemsChangedEventHandler<RangeDataRow>(Rows_ItemsAdded);
      rows.ItemsRemoved += new CollectionItemsChangedEventHandler<RangeDataRow>(Rows_ItemsRemoved);
      rows.ItemsReplaced += new CollectionItemsChangedEventHandler<RangeDataRow>(Rows_ItemsReplaced);
      rows.CollectionReset += new CollectionItemsChangedEventHandler<RangeDataRow>(Rows_CollectionReset);
      foreach (var row in Rows) RegisterRowEvents(row);
    }

    protected virtual void Rows_ItemsAdded(object sender, CollectionItemsChangedEventArgs<RangeDataRow> e) {
      foreach (var row in e.Items)
        this.RegisterRowEvents(row);

      this.OnColumnsChanged();
      this.OnColumnNamesChanged();
      this.OnReset();
    }
    protected virtual void Rows_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<RangeDataRow> e) {
      foreach (var row in e.Items)
        this.DeregisterRowEvents(row);

      this.OnColumnsChanged();
      this.OnColumnNamesChanged();
      this.OnReset();
    }
    protected virtual void Rows_ItemsReplaced(object sender, CollectionItemsChangedEventArgs<RangeDataRow> e) {
      foreach (var row in e.OldItems)
        this.DeregisterRowEvents(row);
      foreach (var row in e.Items)
        this.RegisterRowEvents(row);

      this.OnColumnsChanged();
      this.OnColumnNamesChanged();
      this.OnReset();
    }
    protected virtual void Rows_CollectionReset(object sender, CollectionItemsChangedEventArgs<RangeDataRow> e) {
      foreach (var row in e.OldItems)
        this.DeregisterRowEvents(row);
      foreach (var row in e.Items)
        this.RegisterRowEvents(row);

      if (e.OldItems.Count() != e.Items.Count())
        this.OnColumnsChanged();
      this.OnColumnNamesChanged();
      this.OnReset();
    }

    private void RegisterRowEvents(RangeDataRow row) {
      row.Values.ItemsAdded += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsAdded);
      row.Values.ItemsMoved += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsMoved);
      row.Values.ItemsRemoved += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsRemoved);
      row.Values.ItemsReplaced += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsReplaced);
      row.Values.CollectionReset += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_CollectionReset);
      //if (row is DataRow) {
      //  DataRow _row = row as DataRow;
      //  _row.Values.ItemsAdded += new CollectionItemsChangedEventHandler<IndexedItem<double>>(Values_ItemsAdded);
      //  _row.Values.ItemsMoved += new CollectionItemsChangedEventHandler<IndexedItem<double>>(Values_ItemsMoved);
      //  _row.Values.ItemsRemoved += new CollectionItemsChangedEventHandler<IndexedItem<double>>(Values_ItemsRemoved);
      //  _row.Values.ItemsReplaced += new CollectionItemsChangedEventHandler<IndexedItem<double>>(Values_ItemsReplaced);
      //  _row.Values.CollectionReset += new CollectionItemsChangedEventHandler<IndexedItem<double>>(Values_CollectionReset);
      //} else if(row is RangeDataRow) {
      //  RangeDataRow _row = row as RangeDataRow;
      //  _row.Values.ItemsAdded += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsAdded);
      //  _row.Values.ItemsMoved += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsMoved);
      //  _row.Values.ItemsRemoved += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsRemoved);
      //  _row.Values.ItemsReplaced += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsReplaced);
      //  _row.Values.CollectionReset += new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_CollectionReset);
      //}
    }
    private void DeregisterRowEvents(RangeDataRow row) {
      row.Values.ItemsAdded -= new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsAdded);
      row.Values.ItemsMoved -= new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsMoved);
      row.Values.ItemsRemoved -= new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsRemoved);
      row.Values.ItemsReplaced -= new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_ItemsReplaced);
      row.Values.CollectionReset -= new CollectionItemsChangedEventHandler<IndexedItem<Tuple<double, double>>>(Values_CollectionReset);
    }
    #endregion events

    #region RangeDataRow Value Events
    private void Values_ItemsAdded(object sender, CollectionItemsChangedEventArgs<IndexedItem<Tuple<double, double>>> e) {
      this.OnReset();
    }
    private void Values_ItemsMoved(object sender, CollectionItemsChangedEventArgs<IndexedItem<Tuple<double, double>>> e) {
      this.OnReset();
    }
    private void Values_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<IndexedItem<Tuple<double, double>>> e) {
      this.OnReset();
    }
    private void Values_ItemsReplaced(object sender, CollectionItemsChangedEventArgs<IndexedItem<Tuple<double, double>>> e) {
      this.OnReset();
    }
    private void Values_CollectionReset(object sender, CollectionItemsChangedEventArgs<IndexedItem<Tuple<double, double>>> e) {
      this.OnReset();
    }
    #endregion

    #region IStringConvertibleMatrix Members

    int IStringConvertibleMatrix.Rows {
      get {
        //int maxDR = rows.OfType<DataRow>().Max(r => r.Values.Count);
        //int maxIDR = rows.OfType<RangeDataRow>().Max(r => r.Values.Count);
        //return rows.Count == 0 ? 0 : maxDR >= maxIDR ? maxDR : maxIDR;
        return rows.Count == 0 ? 0 : rows.Max(r => r.Values.Count);
      }
      set { throw new NotSupportedException(); }
    }
    int IStringConvertibleMatrix.Columns {
      get { return rows.Count; }
      set { throw new NotSupportedException(); }
    }
    IEnumerable<string> IStringConvertibleMatrix.ColumnNames {
      get { return rows.Select(r => r.Name); }
      set { throw new NotSupportedException(); }
    }
    IEnumerable<string> IStringConvertibleMatrix.RowNames {
      get { return Enumerable.Empty<string>(); }
      set { throw new NotSupportedException(); }
    }

    bool IStringConvertibleMatrix.SortableView {
      get { return true; }
      set { throw new NotSupportedException(); }
    }
    bool IStringConvertibleMatrix.ReadOnly {
      get { return true; }
    }

    string IStringConvertibleMatrix.GetValue(int rowIndex, int columnIndex) {
      if (columnIndex < rows.Count) {
        string columnName = ((IStringConvertibleMatrix)this).ColumnNames.ElementAt(columnIndex);
        if(rows.ContainsKey(columnName)) {
          RangeDataRow row = rows[columnName];
          if(rowIndex < row.Values.Count) return "[" + row.Values[rowIndex].Item1 + "," + row.Values[rowIndex].Item1 + "]";
        }
        //if(rows.ContainsKey(columnName)) {
        //  if(rows[columnName] is DataRow) {
        //    DataRow row = rows[columnName] as DataRow;
        //    if (rowIndex < row.Values.Count) return row.Values[rowIndex].ToString();
        //  } else if(rows[columnName] is RangeDataRow) {
        //    RangeDataRow row = rows[columnName] as RangeDataRow;
        //    if (rowIndex < row.Values.Count) return "[" + row.Values[rowIndex].Item1 + "," + row.Values[rowIndex].Item1 + "]";
        //  }
        //}
      }
      return string.Empty;
    }

    bool IStringConvertibleMatrix.Validate(string value, out string errorMessage) {
      throw new NotSupportedException();
    }
    bool IStringConvertibleMatrix.SetValue(string value, int rowIndex, int columnIndex) {
      throw new NotSupportedException();
    }

    public event EventHandler<EventArgs<int, int>> ItemChanged;
    protected virtual void OnItemChanged(int rowIndex, int columnIndex) {
      var handler = ItemChanged;
      if (handler != null) handler(this, new EventArgs<int, int>(rowIndex, columnIndex));
      OnToStringChanged();
    }
    public event EventHandler Reset;
    protected virtual void OnReset() {
      var handler = Reset;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler ColumnsChanged;
    protected virtual void OnColumnsChanged() {
      var handler = ColumnsChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler RowsChanged;
    protected virtual void OnRowsChanged() {
      var handler = RowsChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler ColumnNamesChanged;
    protected virtual void OnColumnNamesChanged() {
      var handler = ColumnNamesChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler RowNamesChanged;
    protected virtual void OnRowNamesChanged() {
      var handler = RowNamesChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler SortableViewChanged;
    protected virtual void OnSortableViewChanged() {
      var handler = SortableViewChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    #endregion
  }
}
